/*
 tpcm -- The main function file that implements the tpcm measurement.
*/

#include "tpcm.h"
#include "shim.h"

#define TRANS(value)                                                                    \
    ((((UINT32)(value) << 24) & 0xFF000000) | (((UINT32)(value) << 8) & 0x00FF0000) |   \
        (((UINT32)(value) >> 8) & 0x0000FF00) | (((UINT32)(value) >> 24) & 0x000000FF))

#define WAIT_TIME_UNIT 200

static EFI_GUID hash2_service_binding_guid = SHIM_EFI_HASH2_SERVICE_BINDING_PROTOCOL_GUID;
static EFI_GUID hash2_guid = SHIM_EFI_HASH2_PROTOCOL_GUID;
static EFI_GUID sm3_guid = SHIM_HASH_ALGORITHM_SM3_GUID;
static EFI_GUID gIpmiInterfaceProtocolGuid = EFI_TPCM_GUID;

static UINT8 oemSignature[OEM_SIG_SIZE] = {0xDB, 0x07, 0x00};

static shim_efi_ipmi_interface_protocol_t *tpcm_ipmi;
static UINT32 bm_stage_base = 1500;

static void tpcm_dump_hex(const CHAR16 *name, void *p, int bytes)
{
    int i = 0;
    unsigned char *data = p;
    int add_newline = 1;

    console_print(L"%s length=%d:\n", name, bytes);
    while (i < bytes) {
        console_print(L"%02x ", (unsigned char)data[i]);
        i++;
        if (i % 16 == 0) {
            console_print(L"\n");
            add_newline = 0;
        } else {
            add_newline = 1;
        }
    }
    if (add_newline) {
        console_print(L"\n");
    }
}

static EFI_STATUS tpcm_get_response_blocked(void)
{
    shim_ipmi_cmd_header request = {IPMI_BMC_LUN, IPMI_NETFN_OEM, IPMI_CMD_GET_MEASURE_PARM};
    OEM_BMC_GET_RESULT_REQUSET get_result_request_data;
    OEM_BMC_GET_RESULT_RESPONSE get_result_response_data;
    UINT16 timeout_ms = SHIM_IPMI_TIMEOUT_MS;
    UINT8 response_length;
    EFI_STATUS efi_status;

    memset(&get_result_request_data, 0, sizeof(get_result_request_data));
    memset(&get_result_response_data, 0, sizeof(get_result_response_data));

    get_result_request_data.OemSignature[0] = oemSignature[0];
    get_result_request_data.OemSignature[1] = oemSignature[1];
    get_result_request_data.OemSignature[OEM_SIG_SIZE - 1] = oemSignature[OEM_SIG_SIZE - 1];
    get_result_request_data.SubCmd = IPMI_SUB_CMD_CONTROL_REQ;
    get_result_request_data.FirmwareType = IPMI_FW_SHIM_GRUB;
    get_result_request_data.FirmwareDetailType = IPMI_FW_DETAIL_GRUB;

    // Polling for 7 seconds
    while (timeout_ms > 0) {
        response_length = sizeof(OEM_BMC_GET_RESULT_RESPONSE);
        msleep(WAIT_TIME_UNIT);
        timeout_ms -= WAIT_TIME_UNIT;

        console_print(L"get result request: request_size[%d], response_length[%d]\n",
            sizeof(get_result_request_data), response_length);
        efi_status = tpcm_ipmi->excute_ipmi_cmd(tpcm_ipmi, request, &get_result_request_data,
            sizeof(get_result_request_data), &get_result_response_data, &response_length, NULL);
        console_print(L"OemSignature iana = [ 0x%X %X %X ], ControlResult = %d\n",
            get_result_response_data.OemSignature[0],
            get_result_response_data.OemSignature[1],
            get_result_response_data.OemSignature[OEM_SIG_SIZE - 1],
            get_result_response_data.ControlResult);
        if (efi_status != EFI_SUCCESS) {
            console_print(L"EFI : attempt to get measurement result failed, ret=%d\n", efi_status);
            continue;
        }
        if (response_length == sizeof(OEM_BMC_GET_RESULT_RESPONSE) &&
            get_result_response_data.ControlResult != IPMI_MEASURE_UNKNOW) {
            console_print(L"ipmi protocol: get tpcm measurement result success\n");
            break;
        }
    }

    if (get_result_response_data.ControlResult == IPMI_MEASURE_SUCCESS) {
        efi_status = EFI_SUCCESS;
    } else {
        efi_status = EFI_INVALID_PARAMETER;
        console_print(L"Error: the tpcm measurement result does not pass, and the startup is rejected");
    }

    return efi_status;
}

static EFI_STATUS tpcm_fillup_hash_content(OEM_BMC_MEASURE_REQUSET *request_data, unsigned char *content,
    CHAR16 *description)
{
    UINT32 filename_len = StrLen(description) * 2 + 1;
    UINT32 stage_base = bm_stage_base++;

    if (filename_len > FIRMWARE_NAME_SIZE) {
        console_print(L"the path strings is pass the size of FirmwareHashContent.uaObj!\n");
        return EFI_INVALID_PARAMETER;
    }

    console_print(L"start filling the hash content.\n");
    request_data->FirmwareHashContent.uiCmdTag = TRANS(TPCM_TAG_REQ_COMMAND);
    request_data->FirmwareHashContent.uiCmdLength = TRANS(sizeof(extern_simple_bmeasure_req_st));
    request_data->FirmwareHashContent.uiCmdCode = TRANS(TPCM_ORD_ExternSimpleBootMeasure);
    request_data->FirmwareHashContent.uiPcr = TRANS(0);

    request_data->FirmwareHashContent.uiStage = TRANS(stage_base);

    memcpy((UINT8 *)(request_data->FirmwareHashContent.uaDigest), content, DEFAULT_HASH_SIZE);
    request_data->FirmwareHashContent.uiObjLen = TRANS(filename_len);

    memcpy((UINT8 *)(request_data->FirmwareHashContent.uaObj), description, filename_len);
    request_data->FirmwareHashContent.uaObj[filename_len - 1] = '\0';

    return EFI_SUCCESS;
}

static EFI_STATUS tpcm_send_request(unsigned char *content, CHAR16 *description)
{
    EFI_STATUS efi_status = EFI_SUCCESS;
    shim_ipmi_cmd_header request = {IPMI_BMC_LUN, IPMI_NETFN_OEM, IPMI_CMD_GET_MEASURE_PARM};
    OEM_BMC_MEASURE_REQUSET request_data;
    OEM_BMC_MEASURE_RESPONSE response_data;
    UINT8 response_length = sizeof(OEM_BMC_MEASURE_RESPONSE);
    UINT32 cmd_len = sizeof(extern_simple_bmeasure_req_st);

    memset(&request_data, 0, sizeof(request_data));
    memset(&response_data, 0, sizeof(response_data));

    request_data.OemSignature[0] = oemSignature[0];
    request_data.OemSignature[1] = oemSignature[1];
    request_data.OemSignature[OEM_SIG_SIZE - 1] = oemSignature[OEM_SIG_SIZE - 1];
    request_data.SubCmd = IPMI_SUB_CMD_MEASURE_REQ;
    request_data.FirmwareType = IPMI_FW_SHIM_GRUB;
    request_data.FirmwareDetailType = IPMI_FW_DETAIL_GRUB;
    request_data.FirmwareHashAlgorithmType = SM3_HASH;
    request_data.FirmwareHashLen = cmd_len;

    /* filling the hash content of request data */
    efi_status = tpcm_fillup_hash_content(&request_data, content, description);
    if (efi_status != EFI_SUCCESS) {
        goto out;
    }
    console_print(L"sizeof(request_data)=%d\n", sizeof(request_data));

    /* send the hash request to tpcm chips by ipmi */
    efi_status = tpcm_ipmi->excute_ipmi_cmd(tpcm_ipmi, request, &request_data, (UINT8)sizeof(OEM_BMC_MEASURE_REQUSET),
                                            &response_data, &response_length, NULL);
    if (efi_status != EFI_SUCCESS) {
        console_print(L"ipmi protocol: excute_ipmi_cmd send request failed.\n");
        goto out;
    }
    console_print(L"ipmi protocol: send tpcm measure request success\n");

out:
    return efi_status;
}

static EFI_HANDLE tpcm_efi_service_binding(EFI_GUID *service_binding_guid, EFI_HANDLE image_handle)
{
    EFI_STATUS efi_status;
    EFI_SERVICE_BINDING *service = NULL;
    EFI_HANDLE child_dev = NULL;

    efi_status = LibLocateProtocol(service_binding_guid, (VOID **)&service);
    if (EFI_ERROR(efi_status)) {
        console_print(L"LibLocateProtocol failed\n");
        return NULL;
    }
    if (!service) {
        console_print(L"couldn't open efi service binding protocol\n");
        return NULL;
    }
    efi_status = service->CreateChild(service, &child_dev);
    if (EFI_ERROR(efi_status)) {
        console_print(L"Failed to create child device of http service %x\n", efi_status);
        return NULL;
    }

    return child_dev;
}

static EFI_STATUS tpcm_efi_hash2(shim_efi_hash2_protocol_t *hash2, unsigned char *buf, size_t size,
    unsigned char *output)
{
    EFI_STATUS efi_status;

    if (!hash2->hash_init || !hash2->hash_update || !hash2->hash_final) {
        efi_status = EFI_INVALID_PARAMETER;
        console_print(L"the functions of hash2 has NULL!\n");
        goto out;
    }
    efi_status = hash2->hash_init(hash2, &sm3_guid);
    if (efi_status != EFI_SUCCESS) {
        console_print(L"hash_init failed.\n");
        goto out;
    }
    efi_status = hash2->hash_update(hash2, buf, size);
    if (efi_status != EFI_SUCCESS) {
        console_print(L"hash_update failed.\n");
        goto out;
    }
    efi_status = hash2->hash_final(hash2, (shim_efi_hash2_output *)output);
    if (efi_status != EFI_SUCCESS) {
        console_print(L"hash_final failed.\n");
        goto out;
    }

out:
    return efi_status;
}

static EFI_STATUS tpcm_get_hash(unsigned char *buf, size_t size, EFI_HANDLE image_handle, unsigned char *content)
{
    EFI_STATUS efi_status = EFI_SUCCESS;
    shim_efi_hash2_protocol_t *hash2 = NULL;
    EFI_HANDLE hash_handle = NULL;
    unsigned char output[DEFAULT_HASH_SIZE] = {0};

    console_print(L"tpcm_get_hash start binding service.\n");
    hash_handle = tpcm_efi_service_binding(&hash2_service_binding_guid, image_handle);
    if (!hash_handle) {
        console_print(L"hash2 service binding failed.\n");
        efi_status = EFI_INVALID_PARAMETER;
        goto out;
    }
    console_print(L"tpcm_get_hash binding service success.\n");

    efi_status = gBS->OpenProtocol(
        hash_handle, &hash2_guid, (VOID **)&hash2, image_handle, NULL, EFI_OPEN_PROTOCOL_GET_PROTOCOL);
    if (EFI_ERROR(efi_status)) {
        console_print(L"tpcm_get_hash: gBS->OpenProtocol fail\n");
        goto out;
    }
    if (!hash2) {
        console_print(L"hash2 protocol open failed.\n");
        efi_status = EFI_INVALID_PARAMETER;
        goto out;
    }
    console_print(L"tpcm_get_hash get protocol success.\n");

    efi_status = tpcm_efi_hash2(hash2, buf, size, output);
    if (efi_status != EFI_SUCCESS) {
        console_print(L"tpcm_efi_hash2 failed.\n");
        goto out;
    }

    tpcm_dump_hex(L"tpcm BIOS hash output: ", output, DEFAULT_HASH_SIZE);
    memcpy(content, output, DEFAULT_HASH_SIZE);

out:
    return efi_status;
}

static EFI_STATUS tpcm_do_measure(unsigned char *buf, size_t size, CHAR16 *description, EFI_HANDLE image_handle)
{
    EFI_STATUS efi_status;
    unsigned char content[DEFAULT_HASH_SIZE] = {0};

    /* step1: get the hash of grub */
    efi_status = tpcm_get_hash(buf, size, image_handle, content);
    if (efi_status != EFI_SUCCESS) {
        console_print(L"get firmware hash content failed\n");
        goto out;
    }

    /* step2: send the measure request to tpcm */
    efi_status = tpcm_send_request(content, description);
    if (efi_status != EFI_SUCCESS) {
        console_print(L"tpcm_send_request send request failed\n");
        goto out;
    }

    /* step3: get the result of measure request */
    efi_status = tpcm_get_response_blocked();
    if (efi_status != EFI_SUCCESS) {
        console_print(L"tpcm_get_response_blocked get result failed\n");
        goto out;
    }

out:
    return efi_status;
}

static BOOLEAN tpcm_get_switch(void)
{
    UINT8 response_length;
    EFI_STATUS efi_status = EFI_SUCCESS;
    shim_ipmi_cmd_header request = {IPMI_BMC_LUN, IPMI_NETFN_OEM, IPMI_CMD_GET_MEASURE_PARM};
    OEM_BMC_GET_RESULT_REQUSET get_tpcm_request_value;
    OEM_BMC_GET_RESULT_RESPONSE get_tpcm_response_value;

    memset(&get_tpcm_request_value, 0, sizeof(get_tpcm_request_value));
    memset(&get_tpcm_response_value, 0, sizeof(get_tpcm_response_value));

    get_tpcm_request_value.OemSignature[0] = oemSignature[0];
    get_tpcm_request_value.OemSignature[1] = oemSignature[1];
    get_tpcm_request_value.OemSignature[OEM_SIG_SIZE - 1] = oemSignature[OEM_SIG_SIZE - 1];
    get_tpcm_request_value.SubCmd = IPMI_SUB_CMD_SWITCH_REQ;
    get_tpcm_request_value.FirmwareType = IPMI_FW_OS;
    get_tpcm_request_value.FirmwareDetailType = IPMI_FW_DETAIL_GRUB;
    response_length = sizeof(OEM_BMC_GET_RESULT_RESPONSE);

    if (!tpcm_ipmi->excute_ipmi_cmd) {
        console_print(L"tpcm_ipmi->excute_ipmi_cmd is NULL, some error may occur below shim!\n");
        return FALSE;
    }
    efi_status = tpcm_ipmi->excute_ipmi_cmd(tpcm_ipmi, request, &get_tpcm_request_value, sizeof(get_tpcm_request_value),
                                            &get_tpcm_response_value, &response_length, NULL);
    if (efi_status != EFI_SUCCESS) {
        console_print(L"ipmi get tpcm switch failed.\n");
        return FALSE;
    }

    if (get_tpcm_response_value.ControlResult != IPMI_SWITCH_OPEN) {
        console_print(L"tpcm switch close, skip measure.\n");
        return FALSE;
    }

    return TRUE;
}

static EFI_STATUS tpcm_check_ipmi(void)
{
    EFI_STATUS efi_status;

    efi_status = LibLocateProtocol(&gIpmiInterfaceProtocolGuid, (VOID **)&tpcm_ipmi);
    if (EFI_ERROR(efi_status)) {
        console_print(L"tpcm is not support.");
        return EFI_INVALID_PARAMETER;
    }
    if (!tpcm_ipmi) {
        console_print(L"Error: tpcm_ipmi is NULL");
        return EFI_INVALID_PARAMETER;
    }

    return EFI_SUCCESS;
}

static EFI_STATUS tpcm_ipmi_measure(unsigned char *buf, size_t size, void *description, EFI_HANDLE image_handle)
{
    EFI_STATUS efi_status;
    BOOLEAN switch_flag = FALSE;

    /* step1: check if the tpcm chips is existed. */
    efi_status = tpcm_check_ipmi();
    if (EFI_ERROR(efi_status)) {
        return EFI_SUCCESS;
    }

    /* step2: check if the tpcm switch is on. */
    switch_flag = tpcm_get_switch();

    /* step3: do measure if the tpcm switch is on. */
    if (switch_flag == TRUE) {
        return tpcm_do_measure(buf, size, description, image_handle);
    }

    return EFI_SUCCESS;
}

EFI_STATUS
tpcm_measure_grub(void *context, unsigned char *buf, size_t size, EFI_HANDLE image_handle)
{
    if (context == NULL || buf == NULL || size <= 0) {
        perror(L"the parameter passed to tpcm_measure_grub is error!\n");
        return EFI_INVALID_PARAMETER;
    }

    if (StrCmp((CHAR16 *)context, DEFAULT_LOADER)) {
        console_print(L"only grub is measured by tpcm, not for FALLBACK and MOK_MANAGER.\n");
        return EFI_SUCCESS;
    }

    return tpcm_ipmi_measure(buf, size, context, image_handle);
}